/********************************************************************************************************************/
--
-- Converts projected point on a 3d object to UV coords
-- (c) Gabriel Coriiu
--
/********************************************************************************************************************/
(
	local TheTargetObj = undefined
	local MapChannel = 1 -- usually you want to keep this 2 (lightmap)
	local bRealTimeUpdate = false
	local ProjectionDirection = point3 1 0 0
	local PointOffset = 0.25
	local bGetFromView = false
	local bGetFromVertNRM = false
	--====================================================================================================================
	fn getArrayOfMapVerts2VertsPOLY theObj theMapChannel =
	(
		numMapVerts = polyOp.getNumMapVerts theObj theMapChannel
		mapVerts2Verts = for mv=1 to numMapVerts collect #()
		numMapFaces = polyOp.getNumMapFaces theObj theMapChannel
		for f = 1 to numMapFaces do (
			theMapFace = polyOp.getMapFace theObj theMapChannel f
			polyFace = polyOp.getFaceVerts theObj f
			for mv=1 to theMapFace.count do (
				mapVert = theMapFace[mv]
				if findItem mapVerts2Verts[mapVert] polyFace[mv] == 0 do append mapVerts2Verts[mapVert] polyFace[mv]
				)
			)
		mapVerts2Verts
	)

	fn getArrayOfMapVerts2VertsMESH theObj theMapChannel =
	(
		numMapVerts = meshOp.getNumMapVerts theObj theMapChannel
		mapVerts2Verts = for mv=1 to numMapVerts collect #()
		numMapFaces = meshOp.getNumMapFaces theObj theMapChannel
		for f = 1 to numMapFaces do (
			theMapFace = meshOp.getMapFace theObj theMapChannel f
			theMapFace = #(theMapFace.x as integer,theMapFace.y as integer,theMapFace.z as integer)
			meshFace = getFace theObj f
			meshFace = #(meshFace.x as integer,meshFace.y as integer,meshFace.z as integer)
			for mv=1 to theMapFace.count do (
				mapVert = theMapFace[mv]
				if findItem mapVerts2Verts[mapVert] meshFace[mv] == 0 do append mapVerts2Verts[mapVert] meshFace[mv]
				)
			)
		mapVerts2Verts
	)

	fn getArrayOfMapVerts2Verts theObj theMapChannel =
	(
		array=false
		case classOf theObj of
			(
			Editable_mesh: array=getArrayOfMapVerts2VertsMESH theObj theMapChannel
			Editable_Poly: array=getArrayOfMapVerts2VertsPOLY theObj theMapChannel
			)
		array
	)

	fn getArrayOfVerts2MapVerts mapVerts2Verts =
	(
		verts2MapVerts=#()
		for mv=1 to mapVerts2Verts.count do
			(
			currentVerts=mapVerts2Verts[mv]
			if currentVerts!=undefined do
				(
				for v=1 to currentVerts.count do
					(
				if verts2MapVerts[currentVerts[v]] == undefined
					then verts2MapVerts[currentVerts[v]]=#(mv)
					else append verts2MapVerts[currentVerts[v]] mv
					)
				)
			)
		verts2MapVerts
	)
	--====================================================================================================================

	fn drawLineBetweenTwoPoints pointA pointB =
	(
	  ss = SplineShape pos:pointA
	  addNewSpline ss
	  addKnot ss 1 #corner #line PointA
	  addKnot ss 1 #corner #line PointB
	  updateShape ss
	  ss
	)
	
	function GetUVcoordsOfIntersectedPoint TargetObj ThePoint =
	(
		local tf, tv, tv1, tv2, tv3, arr
		local r
		
		if ( ClassOf TargetObj != Editable_Mesh ) then
			ConvertToMesh TargetObj
		
		r = ray ThePoint (ProjectionDirection * 10)

		-- Get the Intersection details
		arr = (intersectRayEx TargetObj r)		
		if ( arr == undefined ) then
		(
			r = ray (ThePoint - ProjectionDirection*PointOffset) ProjectionDirection
			arr = (intersectRayEx TargetObj r)
		)
		
		if ( arr != undefined ) then
		(
			-- Get the texture face
			tf = getTVFace TargetObj arr[2]

			-- Get the UVW verts of the face
			tv1 = getTVert TargetObj tf.x
			tv2 = getTVert TargetObj tf.y
			tv3 = getTVert TargetObj tf.z

			-- Calculate the texture vertices at point of intersection from
			-- the barycentric coordinates
			tv = tv1*arr[3].x + tv2*arr[3].y + tv3*arr[3].z
		)
		else
		(
			tv = point3 0 0 0
			--format "arr is undefined\n"
		)
		
		return tv
	)

	function getViewDirection =
	(
		-- The affine TM transforms from world coords to view coords -- so we need the inverse of this matrix
		local coordSysTM = Inverse(getViewTM())
		-- The Z axis of this matrix is the view direction.
		local viewDir = -coordSysTM.row3
		-- get the view position from this matrix
		local viewPt = coordSysTM.row4
		return (normalize viewDir)
	)

	function ProjectUVs &load_bar_color &load_bar_value = 
	(
		local Counter = 0
		
		if ( bGetFromView ) then
			ProjectionDirection = getViewDirection()
		
		if ( TheTargetObj == undefined ) then
			messageBox ("TheTargetObj is undefined \n")
		else if ( not meshop.getMapSupport TheTargetObj.mesh MapChannel ) then
			messageBox ("UV channel " + (MapChannel as string) + " does not exist for object \"" + TheTargetObj.name + "\"")
		else if ( TheTargetObj != selection[1] ) then
		(
			local auxMapChannel = 3
			
			convertToMesh TheTargetObj
			
			for ch = 1 to 50 do
			(
				if ( not meshop.getMapSupport TheTargetObj ch ) then
				(
					-- find aux map channel to copy ch1 to, because we need to copy ch2 to ch1 in order for this to work
					auxMapChannel = ch
					break
				)
			)
			
			try (
				channelinfo.copychannel TheTargetObj 3 1
				channelinfo.pastechannel TheTargetObj 3 auxMapChannel
				
				channelinfo.copychannel TheTargetObj 3 MapChannel
				channelinfo.pastechannel TheTargetObj 3 1
			)
			catch()
			
			for obj in selection where (obj != TheTargetObj and TheTargetObj != undefined and classOf obj == Editable_Poly) do
			(
				polyop.setMapSupport obj MapChannel true
				polyop.applyUVWMap obj #planar
				obj.EditablePoly.deleteIsoVerts()
				
				local UVverts = (getArrayOfVerts2MapVerts (getArrayOfMapVerts2Verts obj MapChannel))
				
				local meshSnapshot = snapshotAsMesh obj
				for i = 1 to obj.numVerts do
				(
					if (bGetFromVertNRM) then
						ProjectionDirection = -(getNormal meshSnapshot i)
					
					local VertPos = polyOp.getVert obj i					
					local UVWcoord = (point3 1 0 0) + (GetUVcoordsOfIntersectedPoint TheTargetObj VertPos)
					
					-- debug lines
					-- drawLineBetweenTwoPoints VertPos (VertPos - ProjectionDirection * 10)
					
					for j = 1 to UVverts[i].count do
					(
						--print UVWcoord
						polyOp.setMapVert obj MapChannel UVverts[i][j] UVWcoord
					)
				)
				delete meshSnapshot
				
				Counter = Counter + 1
				load_bar_color = color 0 0 255
				load_bar_value = (Counter as float)*100.0/(selection.count as float)
				
		-- 			update obj
		-- 			if ( not bRealTimeUpdate ) then
		-- 				convertToPoly obj					
			)
			load_bar_color = color 0 128 0
			
			try (
			channelinfo.copychannel TheTargetObj 3 auxMapChannel
			channelinfo.pastechannel TheTargetObj 3 1
			channelinfo.clearChannel TheTargetObj auxMapChannel
			)
			catch()
			
			local FilteredSelection = (for obj in selection where (superClassOf obj == GeometryClass and obj != TheTargetObj) collect obj)
			convertToPoly FilteredSelection
			
			if ( not bRealTimeUpdate ) then
			convertToPoly TheTargetObj
		)
	)

	try(destroyDialog  ProjectUVsDialog)catch()
	rollout ProjectUVsDialog "Project UVs"
	(
		--===============================================--
		-- Elements --
		-------------------------------------------------------------
		
		progressbar load_bar width:135 height:7 color:[0,120,255] value:0
		
		spinner ProjOffset "Proj. offset" range:[0,1000,0.25]
		on ProjOffset changed val do
		(
			PointOffset = val
		)
		
		spinner MapChannel_Spinner "Map Channel" range:[-2,1000,1] type:#integer
		on MapChannel_Spinner changed val do
		(
			MapChannel = val
		)
		--getNormal selection[1].mesh v
		dropdownlist ProjDirection "Select projection direction" items:#("    X","    -X","    Y","    -Y","    Z","    -Z","    Get from view","    By vertex Normal") selection:1  width: 135
		on ProjDirection selected i do
		(
			bGetFromView = false
			bGetFromVertNRM = false
			case i of 
			(
				1: ProjectionDirection = point3 1 0 0
				2: ProjectionDirection = point3 -1 0 0
				3: ProjectionDirection = point3 0 1 0
				4: ProjectionDirection = point3 0 -1 0
				5: ProjectionDirection = point3 0 0 1
				6: ProjectionDirection = point3 0 0 -1
				7: bGetFromView = true
				8: bGetFromVertNRM = true
			)
		)
		
		checkbox RealTimeUpdate "Dynamic update" checked:false
		on RealTimeUpdate changed theState do
		bRealTimeUpdate = theState
		
		timer clock "testClock" interval:3000 --tick once a second
		on clock tick do
		(
			if ( bRealTimeUpdate ) then
			(
				ProjectUVs &load_bar.color &load_bar.value
			)
		)
		
		pickbutton chooseit "> Reference Object <" width:130 height:30
		on chooseit picked obj do
		(
			if ( superClassOf obj == GeometryClass ) then
			(
				if ( classOf obj != Editable_Poly and classOf obj != Editable_Mesh ) then
					convertToMesh obj
				
				TheTargetObj = obj
				chooseit.text = "Ref: " + obj.name
			)
		)
		
		button doit "Project" align:#right
		on doit pressed do 
		(
			ProjectUVs &load_bar.color &load_bar.value
		)
		
		hyperLink Seenax "Gabriel Coriiu" color:(color 0 100 255) hoverColor:(color 0 0 255) visitedColor:(color 0 100 255) address:"www.seenax.com" align:#center offset:[-30, -22]
		--===============================================--
	)
	CreateDialog ProjectUVsDialog width:163
)